@@ -4,22 +4,24 @@ pub mod msg;
 mod multitest;
 pub mod state;
 
-pub use error::ContractError;
+use serde::de::DeserializeOwned;
+use serde::Serialize;
 
-use cosmwasm_std::{Addr, BlockInfo, Deps, DepsMut, Env, MessageInfo, Order, StdResult};
-use cw3::{
+pub use error::ContractError;
+use state::{
+    next_id, proposals, Ballot, Config, Proposal, ProposalListResponse, ProposalResponse,
+    TextProposalListResponse, Votes, VotingRules, BALLOTS, BALLOTS_BY_VOTER, CONFIG,
+    TEXT_PROPOSALS,
+};
+
+use cosmwasm_std::{Addr, BlockInfo, Deps, DepsMut, Env, MessageInfo, Order, StdResult, Storage};
+use cw_storage_plus::Bound;
+use cw_utils::maybe_addr;
+use tg3::{
     Status, Vote, VoteInfo, VoteListResponse, VoteResponse, VoterDetail, VoterListResponse,
     VoterResponse,
 };
-use cw_storage_plus::Bound;
-use cw_utils::maybe_addr;
-use serde::de::DeserializeOwned;
-use serde::Serialize;
-use state::{
-    next_id, proposals, Ballot, Config, Proposal, ProposalListResponse, ProposalResponse, Votes,
-    VotingRules, BALLOTS, CONFIG,
-};
-use tg4::Tg4Contract;
+use tg4::{Member, Tg4Contract};
 use tg_bindings::TgradeMsg;
 use tg_utils::Expiration;
 
@@ -47,6 +49,17 @@ pub fn instantiate(
     Ok(Response::default())
 }
 
+fn save_ballot(
+    storage: &mut dyn Storage,
+    proposal_id: u64,
+    sender: &Addr,
+    ballot: Ballot,
+) -> Result<(), ContractError> {
+    BALLOTS.save(storage, (proposal_id, sender), &ballot)?;
+    BALLOTS_BY_VOTER.save(storage, (sender, proposal_id), &ballot)?;
+    Ok(())
+}
+
 pub fn propose<P>(
     deps: DepsMut,
     env: Env,
@@ -61,7 +74,7 @@ where
     let cfg = CONFIG.load(deps.storage)?;
 
     // Only members of the multisig can create a proposal
-    // Additional check if weight >= 1
+    // Additional check if points >= 1
     let vote_power = cfg
         .group_contract
         .is_voting_member(&deps.querier, info.sender.as_str())?;
@@ -74,24 +87,29 @@ where
     let mut prop = Proposal {
         title,
         description,
+        created_by: info.sender.to_string(),
         start_height: env.block.height,
         expires,
         proposal,
         status: Status::Open,
         votes: Votes::yes(vote_power),
         rules: cfg.rules,
-        total_weight: cfg.group_contract.total_weight(&deps.querier)?,
+        total_points: cfg.group_contract.total_points(&deps.querier)?,
     };
     prop.update_status(&env.block);
     let id = next_id(deps.storage)?;
     proposals().save(deps.storage, id, &prop)?;
 
     // add the first yes vote from voter
-    let ballot = Ballot {
-        weight: vote_power,
-        vote: Vote::Yes,
-    };
-    BALLOTS.save(deps.storage, (id, &info.sender), &ballot)?;
+    save_ballot(
+        deps.storage,
+        id,
+        &info.sender,
+        Ballot {
+            points: vote_power,
+            vote: Vote::Yes,
+        },
+    )?;
 
     let resp = msg::ProposalCreationResponse { proposal_id: id };
 
@@ -115,12 +133,10 @@ where
 {
     // ensure proposal exists and can be voted on
     let mut prop = proposals().load(deps.storage, proposal_id)?;
-    if prop.status != Status::Open {
+
+    if prop.current_status(&env.block) != Status::Open {
         return Err(ContractError::NotOpen {});
     }
-    if prop.expires.is_expired(&env.block) {
-        return Err(ContractError::Expired {});
-    }
 
     // use a snapshot of "start of proposal"
     // Must be a member of voting group and have voting power >= 1
@@ -130,13 +146,21 @@ where
             .was_voting_member(&deps.querier, &info.sender, prop.start_height)?;
 
     // cast vote if no vote previously cast
-    BALLOTS.update(deps.storage, (proposal_id, &info.sender), |bal| match bal {
-        Some(_) => Err(ContractError::AlreadyVoted {}),
-        None => Ok(Ballot {
-            weight: vote_power,
+    if BALLOTS
+        .may_load(deps.storage, (proposal_id, &info.sender))?
+        .is_some()
+    {
+        return Err(ContractError::AlreadyVoted {});
+    }
+    save_ballot(
+        deps.storage,
+        proposal_id,
+        &info.sender,
+        Ballot {
+            points: vote_power,
             vote,
-        }),
-    })?;
+        },
+    )?;
 
     // update vote tally
     prop.votes.add_vote(vote, vote_power);
@@ -155,14 +179,14 @@ where
 /// as you you wouldn't be able to execute it in the future (If the contract call errors, this status
 /// change will be reverted / ignored).
 pub fn mark_executed<P>(
-    deps: DepsMut,
+    storage: &mut dyn Storage,
     env: Env,
     proposal_id: u64,
 ) -> Result<Proposal<P>, ContractError>
 where
     P: Serialize + DeserializeOwned,
 {
-    let mut proposal = proposals::<P>().load(deps.storage, proposal_id)?;
+    let mut proposal = proposals::<P>().load(storage, proposal_id)?;
 
     // We allow execution even after the proposal "expiration" as long as all votes come in before
     // that point. If it was approved on time, it can be executed any time.
@@ -172,10 +196,19 @@ where
 
     // Set it to executed
     proposal.status = Status::Executed;
-    proposals::<P>().save(deps.storage, proposal_id, &proposal)?;
+    proposals::<P>().save(storage, proposal_id, &proposal)?;
     Ok(proposal)
 }
 
+pub fn execute_text<P>(deps: DepsMut, id: u64, proposal: Proposal<P>) -> Result<(), ContractError>
+where
+    P: Serialize + DeserializeOwned,
+{
+    TEXT_PROPOSALS.save(deps.storage, id, &proposal.into())?;
+
+    Ok(())
+}
+
 pub fn close<P>(
     deps: DepsMut,
     env: Env,
@@ -187,8 +220,15 @@ where
 {
     // anyone can trigger this if the vote passed
 
-    let mut prop = proposals::<P>().load(deps.storage, proposal_id)?;
-    if [Status::Executed, Status::Rejected, Status::Passed]
+    let mut prop = proposals().load(deps.storage, proposal_id)?;
+
+    if prop.status == Status::Rejected {
+        return Err(ContractError::NotOpen {});
+    }
+
+    prop.update_status(&env.block);
+
+    if [Status::Executed, Status::Passed]
         .iter()
         .any(|x| *x == prop.status)
     {
@@ -198,7 +238,6 @@ where
         return Err(ContractError::NotExpired {});
     }
 
-    // set it to failed
     prop.status = Status::Rejected;
     proposals::<P>().save(deps.storage, proposal_id, &prop)?;
 
@@ -225,10 +264,11 @@ where
         title: prop.title,
         description: prop.description,
         proposal: prop.proposal,
+        created_by: prop.created_by,
         status,
         expires: prop.expires,
         rules,
-        total_weight: prop.total_weight,
+        total_points: prop.total_points,
         votes: prop.votes,
     })
 }
@@ -244,10 +284,11 @@ fn map_proposal<P>(
         title: prop.title,
         description: prop.description,
         proposal: prop.proposal,
+        created_by: prop.created_by,
         status,
         expires: prop.expires,
         rules: prop.rules,
-        total_weight: prop.total_weight,
+        total_points: prop.total_points,
         votes: prop.votes,
     })
 }
@@ -271,6 +312,21 @@ where
     Ok(ProposalListResponse { proposals: props? })
 }
 
+pub fn list_text_proposals(
+    deps: Deps,
+    start_after: Option<u64>,
+    limit: usize,
+) -> StdResult<TextProposalListResponse> {
+    let start = start_after.map(Bound::exclusive_int);
+    let props: StdResult<Vec<_>> = TEXT_PROPOSALS
+        .range(deps.storage, start, None, Order::Ascending)
+        .take(limit)
+        .map(|r| r.map(|(_, p)| p))
+        .collect();
+
+    Ok(TextProposalListResponse { proposals: props? })
+}
+
 pub fn reverse_proposals<P>(
     deps: Deps,
     env: Env,
@@ -294,9 +350,10 @@ pub fn query_vote(deps: Deps, proposal_id: u64, voter: String) -> StdResult<Vote
     let voter_addr = deps.api.addr_validate(&voter)?;
     let prop = BALLOTS.may_load(deps.storage, (proposal_id, &voter_addr))?;
     let vote = prop.map(|b| VoteInfo {
+        proposal_id,
         voter,
         vote: b.vote,
-        weight: b.weight,
+        points: b.points,
     });
     Ok(VoteResponse { vote })
 }
@@ -317,9 +374,37 @@ pub fn list_votes(
         .map(|item| {
             let (voter, ballot) = item?;
             Ok(VoteInfo {
+                proposal_id,
                 voter: voter.into(),
                 vote: ballot.vote,
-                weight: ballot.weight,
+                points: ballot.points,
+            })
+        })
+        .collect();
+
+    Ok(VoteListResponse { votes: votes? })
+}
+
+pub fn list_votes_by_voter(
+    deps: Deps,
+    voter: String,
+    start_after: Option<u64>,
+    limit: usize,
+) -> StdResult<VoteListResponse> {
+    let start = start_after.map(Bound::exclusive_int);
+    let voter_addr = deps.api.addr_validate(&voter)?;
+
+    let votes: StdResult<Vec<_>> = BALLOTS_BY_VOTER
+        .prefix(&voter_addr)
+        .range(deps.storage, start, None, Order::Ascending)
+        .take(limit)
+        .map(|item| {
+            let (proposal_id, ballot) = item?;
+            Ok(VoteInfo {
+                proposal_id,
+                voter: voter.clone(),
+                vote: ballot.vote,
+                points: ballot.points,
             })
         })
         .collect();
@@ -330,9 +415,9 @@ pub fn list_votes(
 pub fn query_voter(deps: Deps, voter: String) -> StdResult<VoterResponse> {
     let cfg = CONFIG.load(deps.storage)?;
     let voter_addr = deps.api.addr_validate(&voter)?;
-    let weight = cfg.group_contract.is_member(&deps.querier, &voter_addr)?;
+    let points = cfg.group_contract.is_member(&deps.querier, &voter_addr)?;
 
-    Ok(VoterResponse { weight })
+    Ok(VoterResponse { points })
 }
 
 pub fn list_voters(
@@ -345,10 +430,7 @@ pub fn list_voters(
         .group_contract
         .list_members(&deps.querier, start_after, limit)?
         .into_iter()
-        .map(|member| VoterDetail {
-            addr: member.addr,
-            weight: member.weight,
-        })
+        .map(|Member { addr, points }| VoterDetail { addr, points })
         .collect();
     Ok(VoterListResponse { voters })
 }
